#include "MatchFinder.h"
/* memcpy must be inlined - we do not want to use RTL */
#include <memory.h>
#pragma intrinsic(memcpy)

#define kMaxValForNormalize (((unsigned)1<<31)-1)

/* settings for bt4:
	defined HASH_ARRAY_2
	defined HASH_ARRAY_3
*/

#define kHash2Size 0x400
#define kNumHashDirectBytes 0
#define kNumHashBytes 4
#define kHash3Size 0x40000
#define kHashSize 0x100000

#define kHashSizeSum (kHashSize+kHash2Size+kHash3Size)
#define kHash2Offset kHashSize
#define kHash3Offset (kHashSize+kHash2Size)

static unsigned _cyclicBufferPos;
static unsigned _cyclicBufferSize;
static unsigned _matchMaxLen;
static unsigned* _hash;
static unsigned _cutValue;

static byte* _bufferBase;
static unsigned _posLimit;
static bool _streamEndWasReached;
static byte* _pointerToLastSafePosition;
static byte* _buffer;
static unsigned _blockSize;
static unsigned _pos;
static unsigned _keepSizeBefore;
static unsigned _keepSizeAfter;
static unsigned _keepSizeReserv;
static unsigned _streamPos;

/* LZ Window */

static void LZInWindow_Create(unsigned keepSizeBefore,unsigned keepSizeAfter,unsigned keepSizeReserv,byte**mem)
{
	_keepSizeBefore = keepSizeBefore;
	_keepSizeAfter = keepSizeAfter;
	_keepSizeReserv = keepSizeReserv;
	_blockSize = keepSizeBefore + keepSizeAfter + keepSizeReserv;
	_bufferBase = *mem;
	_blockSize = (_blockSize + 3) & ~3;
	*mem += _blockSize;
	_pointerToLastSafePosition = _bufferBase + _blockSize - keepSizeAfter;
}

static void ReadBlock(void)
{
	if (_streamEndWasReached)
		return;
	for (;;)
	{
		unsigned size;
		size = (unsigned)(_bufferBase-_buffer) + _blockSize - _streamPos;
		if (!size) return;
		if (size > pack_length - pack_pos)
			size = pack_length - pack_pos;
		memcpy(_buffer+_streamPos,curin,size);
		curin += size;
		pack_pos += size;
		if (size == 0)
		{
			byte* pointerToPosition;
			_posLimit = _streamPos;
			pointerToPosition = _buffer + _posLimit;
			if (pointerToPosition > _pointerToLastSafePosition)
				_posLimit = _pointerToLastSafePosition - _buffer;
			_streamEndWasReached = true;
			return;
		}
		_streamPos += size;
		if (_streamPos >= _pos + _keepSizeAfter)
		{
			_posLimit = _streamPos - _keepSizeAfter;
			return;
		}
	}
}

static void LZInWindow_Init(void)
{
	_buffer = _bufferBase;
	_pos = 0;
	_streamPos = 0;
	_streamEndWasReached = false;
	ReadBlock();
}

const byte* GetPointerToCurrentPos(void) {return _buffer+_pos;}

static void MoveBlock(void)
{
	unsigned offset,numBytes;
	byte* sou;
	offset = _buffer-_bufferBase+_pos-_keepSizeBefore;
	numBytes = _buffer-_bufferBase+_streamPos-offset;
	/* memmove(_bufferBase,_bufferBase+offset,numBytes); */
	sou = _bufferBase+offset;
	if (_bufferBase <= sou)
		memcpy(_bufferBase,sou,numBytes);
	else
	{
		unsigned i;
		for (i=numBytes;i--;)
			_bufferBase[i] = sou[i];
	}
	_buffer -= offset;
}

static void LZInWindow_MovePos(void)
{
	_pos++;
	if (_pos > _posLimit)
	{
		const byte* pointerToPosition = _buffer+_pos;
		if (pointerToPosition > _pointerToLastSafePosition)
			MoveBlock();
		ReadBlock();
	}
}

byte GetIndexByte(int index) {return _buffer[_pos+index];}

unsigned GetMatchLen(int index,unsigned distance,unsigned limit)
{
	byte* pby;
	unsigned i;
	if (_streamEndWasReached)
		if ((_pos+index)+limit > _streamPos)
			limit = _streamPos - (_pos+index);
	distance++;
	pby = _buffer + _pos + index;
	for (i=0;i<limit && pby[i]==pby[(int)(i-distance)];i++) ;
	return i;
}

unsigned GetNumAvailableBytes(void) {return _streamPos-_pos;}

void ReduceOffsets(int subValue)
{
	_buffer += subValue;
	_posLimit -= subValue;
	_pos -= subValue;
	_streamPos -= subValue;
}

/* Binary tree Match Finder */

static unsigned crc_table[256];

void MatchFinder_Create(unsigned historySize,unsigned keepAddBufferBefore,unsigned matchMaxLen,unsigned keepAddBufferAfter,byte**mem)
{
	unsigned sizeReserv;
	sizeReserv = (historySize + keepAddBufferBefore + matchMaxLen + keepAddBufferAfter)/2+256;
	LZInWindow_Create(historySize+keepAddBufferBefore,matchMaxLen+keepAddBufferAfter,sizeReserv,mem);
	_matchMaxLen = matchMaxLen;
	_cyclicBufferSize = historySize+1;
	_hash = (unsigned*)*mem;
	*mem += (kHashSizeSum + _cyclicBufferSize*2) * sizeof(unsigned);
	_cutValue = 0xFF;
}

void MatchFinder_Init(void)
{
	unsigned i,j,r;
	LZInWindow_Init();
	for (i=0;i<kHashSizeSum;i++)
		_hash[i] = 0;
	_cyclicBufferPos = 0;
	ReduceOffsets(-1);
	for (i=0;i<256;i++)
	{
		r = i;
		for (j=0;j<8;j++)
			if (r & 1)
				r = (r>>1) ^ 0xEDB88320;
			else
				r >>= 1;
		crc_table[i] = r;
	}
}

static unsigned Hash(const byte* ptr, unsigned* hash2Value, unsigned* hash3Value)
{
	unsigned temp;
	temp = crc_table[ptr[0]] ^ ptr[1];
	*hash2Value = temp & (kHash2Size - 1);
	*hash3Value = (temp ^ ((unsigned)ptr[2]<<8)) & (kHash3Size - 1);
	return (temp ^ ((unsigned)ptr[2]<<8) ^ (crc_table[ptr[3]] << 5)) & (kHashSize - 1);
}

unsigned GetLongestMatch(unsigned* distances)
{
	unsigned lenLimit,maxLen=0;
	unsigned matchMinPos;
	byte* cur;
	unsigned hash2Value,hash3Value,hashValue;
	unsigned curMatch,curMatch2,curMatch3;
	unsigned *son,*ptr0,*ptr1;
	unsigned len0,len1,count;
	if (_pos + _matchMaxLen <= _streamPos)
		lenLimit = _matchMaxLen;
	else
	{
		lenLimit = _streamPos - _pos;
		if (lenLimit < kNumHashBytes)
			return 0;
	}
	matchMinPos = (_pos>_cyclicBufferSize) ? (_pos-_cyclicBufferSize) : 0;
	cur = _buffer+_pos;
	hashValue = Hash(cur,&hash2Value,&hash3Value);
	curMatch = _hash[hashValue];
	curMatch2 = _hash[kHash2Offset + hash2Value];
	curMatch3 = _hash[kHash3Offset + hash3Value];
	_hash[kHash2Offset + hash2Value] = _pos;
	distances[2] = 0xFFFFFFFF;
	if (curMatch2 > matchMinPos)
		if (_buffer[curMatch2] == cur[0])
		{
			distances[2] = _pos - curMatch2 - 1;
			maxLen = 2;
		}
	_hash[kHash3Offset + hash3Value] = _pos;
	distances[3] = 0xFFFFFFFF;
	if (curMatch3 > matchMinPos)
		if (_buffer[curMatch3] == cur[0])
		{
			distances[3] = _pos - curMatch3 - 1;
			maxLen = 3;
		}
	_hash[hashValue] = _pos;
	son = _hash + kHashSizeSum;
	ptr0 = son + (_cyclicBufferPos << 1) + 1;
	ptr1 = son + (_cyclicBufferPos << 1);
	distances[kNumHashBytes] = 0xFFFFFFFF;
	len0 = len1 = kNumHashDirectBytes;
	count = _cutValue;
	for (;;)
	{
		byte* pb;
		unsigned len,delta;
		unsigned cyclicPos;
		unsigned* pair;
		if (curMatch <= matchMinPos || count--==0)
		{
			*ptr0 = *ptr1 = 0;
			break;
		}
		pb = _buffer+curMatch;
		len = (len0<len1) ? len0 : len1;
		do
			if (pb[len] != cur[len]) break;
		while (++len != lenLimit);
		delta = _pos - curMatch;
		while (maxLen < len)
			distances[++maxLen] = delta-1;
		cyclicPos = (delta <= _cyclicBufferPos) ?
			(_cyclicBufferPos - delta) :
			(_cyclicBufferPos - delta + _cyclicBufferSize);
		pair = son + (cyclicPos<<1);
		if (len != lenLimit)
		{
			if (pb[len] < cur[len])
			{
				*ptr1 = curMatch;
				ptr1 = pair+1;
				curMatch = *ptr1;
				len1 = len;
			}
			else
			{
				*ptr0 = curMatch;
				ptr0 = pair;
				curMatch = *ptr0;
				len0 = len;
			}
		}
		else
		{
			*ptr1 = pair[0];
			*ptr0 = pair[1];
			break;
		}
	}
	if (distances[4] < distances[3])
		distances[3] = distances[4];
	if (distances[3] < distances[2])
		distances[2] = distances[3];
	return maxLen;
}

void DummyLongestMatch(void)
{
	unsigned lenLimit;
	unsigned matchMinPos;
	byte* cur;
	unsigned hash2Value,hash3Value,hashValue;
	unsigned curMatch;
	unsigned* son,*ptr0,*ptr1;
	unsigned len0,len1,count;
	if (_pos + _matchMaxLen <= _streamPos)
		lenLimit = _matchMaxLen;
	else
	{
		lenLimit = _streamPos - _pos;
		if (lenLimit < kNumHashBytes)
			return;
	}
	matchMinPos = (_pos > _cyclicBufferSize) ? (_pos - _cyclicBufferSize) : 0;
	cur = _buffer+_pos;
	hashValue = Hash(cur,&hash2Value,&hash3Value);
	_hash[kHash3Offset + hash3Value] = _pos;
	_hash[kHash2Offset + hash2Value] = _pos;
	curMatch = _hash[hashValue];
	_hash[hashValue] = _pos;
	son = _hash+kHashSizeSum;
	ptr0 = son + (_cyclicBufferPos << 1) + 1;
	ptr1 = son + (_cyclicBufferPos << 1);
	len0 = len1 = kNumHashDirectBytes;
	count = _cutValue;
	for (;;)
	{
		byte* pb;
		unsigned len;
		unsigned delta,cyclicPos;
		unsigned* pair;
		if (curMatch <= matchMinPos || count--==0)
			break;
		pb = _buffer+curMatch;
		len = (len0<len1) ? len0 : len1;
		do
			if (pb[len] != cur[len]) break;
		while (++len != lenLimit);
		delta = _pos - curMatch;
		cyclicPos = (delta <= _cyclicBufferPos) ?
			(_cyclicBufferPos - delta) :
			(_cyclicBufferPos - delta + _cyclicBufferSize);
		pair = son + (cyclicPos << 1);
		if (len != lenLimit)
		{
			if (pb[len] < cur[len])
			{
				*ptr1 = curMatch;
				ptr1 = pair+1;
				curMatch = *ptr1;
				len1 = len;
			}
			else
			{
				*ptr0 = curMatch;
				ptr0 = pair;
				curMatch = *ptr0;
				len0 = len;
			}
		}
		else
		{
			*ptr1 = pair[0];
			*ptr0 = pair[1];
			return;
		}
	}
	*ptr0 = *ptr1 = 0;
}

static void Normalize(void)
{
	unsigned subValue;
	unsigned* items;
	unsigned i,numItems;
	subValue = _pos - _cyclicBufferSize;
	items = _hash;
	numItems = (kHashSizeSum + _cyclicBufferSize*2);
	for (i=0;i<numItems;i++)
	{
		unsigned value;
		value = items[i];
		if (value <= subValue)
			value = 0;
		else
			value -= subValue;
		items[i] = value;
	}
	ReduceOffsets(subValue);
}

void MatchFinder_MovePos(void)
{
	if (++_cyclicBufferPos == _cyclicBufferSize)
		_cyclicBufferPos = 0;
	LZInWindow_MovePos();
	if (_pos == kMaxValForNormalize)
		Normalize();
}
